"""Plot the CPMG profiles."""
import contextlib

import numpy as np
from matplotlib import gridspec as gsp
from matplotlib import pyplot as plt
from matplotlib import ticker
from matplotlib.backends import backend_pdf

from chemex.experiments import plotting


def compute_profiles(data_grouped, params):
    """Compute the CPMG profiles used for plotting."""
    profiles = {}

    for peak, profile in data_grouped.items():
        mask = profile.ncycs != 0
        mask_ref = np.logical_not(mask)

        mag_ref = np.mean(profile.val[mask_ref])

        nu_cpmg = profile.ncycs_to_nu_cpmgs()[mask]

        mag_cal = profile.calculate_profile(params)[mask]
        mag_exp = profile.val[mask]
        mag_err = profile.err[mask]

        # sort values according to increasing nu_cpmg values
        sorted_items = list(zip(*sorted(zip(nu_cpmg, mag_cal, mag_exp, mag_err))))
        nu_cpmg, mag_cal, mag_exp, mag_err = sorted_items

        mag_ens = mag_exp + mag_err * np.random.randn(1000, len(mag_exp))

        r2_cal = -np.log(mag_cal / mag_ref) / profile.exp_details["time_t2"]
        r2_exp = -np.log(mag_exp / mag_ref) / profile.exp_details["time_t2"]
        r2_ens = -np.log(mag_ens / mag_ref) / profile.exp_details["time_t2"]

        r2_erd, r2_eru = abs(np.percentile(r2_ens, [15.9, 84.1], axis=0) - r2_exp)

        profiles[peak] = nu_cpmg, r2_cal, r2_exp, r2_erd, r2_eru

    return profiles


def write_profile_fit(name, profile, file_txt):
    """Write the fitted CPMG profile."""

    file_txt.write("[{}]\n".format(name.upper()))

    file_txt.write("# {:>17s}   {:>17s}\n".format("NU_CPMG", "R2"))

    for nu_cpmg, r2_cal, _r2_exp, _r2_erd, _r2_eru in sorted(zip(*profile)):
        file_txt.write(f"  {nu_cpmg:17.8e} = {r2_cal:17.8e}\n")

    file_txt.write("\n")


def write_profile_exp(name, profile, file_txt):
    """Write the experimental CPMG profile."""

    file_txt.write("[{}]\n".format(name.upper()))

    file_txt.write(
        "# {:>17s}   {:>17s} {:>17s} {:>17s}\n".format(
            "NU_CPMG", "R2", "UNCERTAINTY_DOWN", "UNCERTAINTY_UP"
        )
    )

    for nu_cpmg, _r2_cal, r2_exp, r2_erd, r2_eru in sorted(zip(*profile)):
        file_txt.write(
            "  {:17.8e} = {:17.8e} {:17.8e} {:17.8e}\n".format(
                nu_cpmg, r2_exp, r2_erd, r2_eru
            )
        )

    file_txt.write("\n")


def plot_data(data, params, output_dir):
    """Write experimental and fitted data to a file and plot the CPMG profiles.

    - *.fit: contains the experimental and fitted data
    - *.pdf: contains the plot of experimental and fitted data

    """
    datasets = dict()

    for data_point in data:
        experiment_name = data_point.experiment_name
        datasets.setdefault(experiment_name, []).append(data_point)

    for experiment_name, dataset in datasets.items():
        basename = output_dir / experiment_name
        name_pdf = basename.with_suffix(".pdf")
        name_fit = basename.with_suffix(".fit")
        name_exp = basename.with_suffix(".exp")

        print((f"  * {name_pdf} [.fit, .exp]"))

        data_grouped = plotting.group_data(dataset)
        profiles = compute_profiles(data_grouped, params)

        with contextlib.ExitStack() as stack:
            file_pdf = stack.enter_context(backend_pdf.PdfPages(name_pdf))
            file_fit = stack.enter_context(name_fit.open("w"))
            file_exp = stack.enter_context(name_exp.open("w"))

            for peak in sorted(profiles):
                nu_cpmg, r2_cal, r2_exp, r2_erd, r2_eru = profiles[peak]
                write_profile_exp(peak.assignment, profiles[peak], file_exp)
                write_profile_fit(peak.assignment, profiles[peak], file_fit)

                r2_min = min(min(r2_cal), min(r2_exp - r2_erd))
                r2_max = max(max(r2_cal), max(r2_exp + r2_eru))
                ymin, ymax = plotting.set_lim([r2_min, r2_max], 0.10)

                # Matplotlib #
                grid_spec = gsp.GridSpec(2, 1, height_ratios=[1, 4])

                ax1 = plt.subplot(grid_spec[0])
                ax2 = plt.subplot(grid_spec[1])

                ax1.axhline(0, color=plotting.PALETTE["Black"]["Text"], linewidth=0.5)
                ax2.axhline(0, color=plotting.PALETTE["Black"]["Text"], linewidth=0.5)

                ########################

                ax2.plot(
                    nu_cpmg,
                    r2_cal,
                    linestyle="-",
                    color=plotting.PALETTE["Grey"]["700"],
                    zorder=2,
                )

                ax2.errorbar(
                    nu_cpmg,
                    r2_exp,
                    yerr=(r2_erd, r2_eru),
                    fmt="o",
                    markeredgecolor=plotting.PALETTE["Red"]["500"],
                    ecolor=plotting.PALETTE["Red"]["500"],
                    markerfacecolor="None",
                    zorder=3,
                )

                xmin, xmax = plotting.set_lim(nu_cpmg, 0.10)

                ax2.set_xlim(xmin, xmax)
                ax2.set_ylim(ymin, ymax)

                ax2.xaxis.set_major_locator(ticker.MaxNLocator(6))
                ax2.yaxis.set_major_locator(ticker.MaxNLocator(6))

                ax2.set_xlabel(r"$\nu_{CPMG} \ (Hz)$")
                ax2.set_ylabel(r"$R_{2,eff} \ (s^{-1})$")

                ax1.set_title("{:s}".format(peak.assignment.upper()))

                ax2.yaxis.set_ticks_position("left")
                ax2.xaxis.set_ticks_position("bottom")

                ########################

                deltas = np.asarray(r2_exp) - np.asarray(r2_cal)

                ax1.errorbar(
                    nu_cpmg,
                    deltas,
                    yerr=(r2_erd, r2_eru),
                    fmt="o",
                    markeredgecolor=plotting.PALETTE["Red"]["500"],
                    ecolor=plotting.PALETTE["Red"]["500"],
                    markerfacecolor="None",
                    zorder=100,
                )

                rmin, _rmax = plotting.set_lim(deltas - r2_erd, 0.2)
                _rmin, rmax = plotting.set_lim(deltas + r2_eru, 0.2)

                ax1.set_xlim(xmin, xmax)
                ax1.set_ylim(rmin, rmax)

                ax1.xaxis.set_major_locator(ticker.MaxNLocator(6))
                ax1.yaxis.set_major_locator(ticker.MaxNLocator(4))

                ax1.xaxis.set_major_formatter(ticker.NullFormatter())

                ax1.ticklabel_format(style="sci", scilimits=(0, 0), axis="y")

                ax1.set_title("{:s}".format(peak.assignment.upper()))
                ax1.set_ylabel("Residual " + r"$(s^{-1})$")

                ########################

                for ax in (ax1, ax2):
                    ax.yaxis.set_ticks_position("left")
                    ax.xaxis.set_ticks_position("bottom")

                ########################

                file_pdf.savefig()
                plt.close()

                ########################

    return



"""The fitting module contains the code for fitting the experimental data."""
import functools
import sys

import lmfit
from scipy import stats

from chemex import datasets
from chemex import parameters
from chemex import util
from chemex.cli import FITMETHODS

ALL_FITMETHODS = lmfit.minimizer.SCALAR_METHODS
ALL_FITMETHODS.update(
    {
        "leastsq": "least-squares using Levenberg-Marquardt",
        "least_squares": "least-squares using Trust Region Reflective algorithm",
        "brute": "grid search using the brute force method",
        "basinhopping": "basinhopping",
        "ampgo": "Adaptive Memory Programming for Global Optimization (AMPGO)",
    }
)
ALLOWED_FITMETHODS = {
    name: desc for name, desc in ALL_FITMETHODS.items() if name in FITMETHODS
}


def run_fit(fit_filename, params, data, cl_fitmethod):
    """Perform the fit."""
    util.header1("Fit")

    fit_config = util.read_cfg_file(fit_filename)

    if not fit_config.sections():
        fit_config.add_section("Standard Calculation")

    for section in fit_config.sections():
        util.header2(section)
        items = fit_config.items(section)
        parameters.set_param_status(params, items)
        clusters = find_independent_clusters(data, params)
        fitmethod = fit_config.get(section, "fitmethod", fallback=cl_fitmethod)

        if fitmethod not in ALLOWED_FITMETHODS.keys():
            exit(
                "The fitting method '{}', as specified in section ['{}'],"
                "is invalid! Please choose from:\n  {}".format(
                    fitmethod, section, list(sorted(ALLOWED_FITMETHODS.keys()))
                )
            )

        print("Fitting method: {}\n".format(ALLOWED_FITMETHODS[fitmethod]))

        for c_name, c_data, c_params in clusters:
            if len(clusters) > 1:
                print(f"[{c_name}]")

            print("Chi2 / Reduced Chi2:")

            c_func = c_data.calculate_residuals
            c_minimizer = lmfit.Minimizer(c_func, c_params)

            try:
                if fitmethod == "brute":
                    c_result = c_minimizer.minimize(method=fitmethod, keep="all")
                else:
                    c_result = c_minimizer.minimize(method=fitmethod)

            except KeyboardInterrupt:
                sys.stderr.write("\n -- Keyboard Interrupt: minimization stopped\n")
                c_result = c_minimizer.minimize(
                    params=c_minimizer.result.params, maxfev=1
                )

            for name, param in c_result.params.items():
                params[name] = param

            print("")

        if len(clusters) > 1:
            c_func = functools.partial(data.calculate_residuals, verbose=False)
            minimizer = lmfit.Minimizer(c_func, params)
            result = minimizer.minimize(maxfev=1)
            result.params = params
        else:
            result = c_result

        print(f"Final Chi2        : {result.chisqr:.3e}")
        print(f"Final Reduced Chi2: {result.redchi:.3e}")

    if result.method != "leastsq":
        print("\nWarning: uncertainties and covariance of fitting parameters are only")
        print("         calculated when using the 'leastsq' fitting method!")

    return result


def find_independent_clusters(data, params):
    """Find clusters of datapoints that depend on disjoint sets of variables.

    For example, if the population of the minor state and the exchange
    rate are set to 'fix', chances are that the fit can be decomposed
    residue-specifically.

    """
    clusters = []

    for profile in data:

        pnames = profile.params.keys()
        pnames_vary = {
            name for name in pnames if params[name].vary and not params[name].expr
        }

        for data_cluster, pnames_cluster, pnames_vary_cluster in clusters:

            if pnames_vary & pnames_vary_cluster:
                data_cluster.append(profile)
                pnames_cluster.extend(pnames)
                pnames_vary_cluster.update(pnames_vary)
                break

        else:

            data_cluster = datasets.DataSet(profile)
            pnames_cluster = list(pnames)
            pnames_vary_cluster = pnames_vary

            clusters.append((data_cluster, pnames_cluster, pnames_vary_cluster))

    clusters_ = []

    for data_cluster, pnames_cluster, pnames_vary_cluster in clusters:

        name_cluster = parameters.ParameterName.from_full_name(
            pnames_vary_cluster.pop()
        )

        while pnames_vary_cluster:
            name_cluster = name_cluster.intersection(
                parameters.ParameterName.from_full_name(pnames_vary_cluster.pop())
            )

        params_cluster = lmfit.Parameters()

        for pname in pnames_cluster:
            param = params[pname]
            params[pname]._delay_asteval = True
            params_cluster[pname] = param

        for param in params_cluster.values():
            param._delay_asteval = False

        clusters_.append((name_cluster, data_cluster, params_cluster))

    return sorted(clusters_)


def write_statistics(result, path):
    """Write fitting statistics to a file."""

    _, ks_p_value = stats.kstest(result.residual, "norm")
    chi2_p_value = 1.0 - stats.chi2.cdf(result.chisqr, result.nfree)

    filename = path / "statistics.fit"

    with open(filename, "w") as f:
        print(f"  * {filename}")

        f.write(f"# Number of data points: {result.ndata}\n")
        f.write(f"# Number of variables: {result.nvarys}\n")
        f.write(f"# Fitting method: {result.method}\n")
        f.write(f"# Number of function evaluations: {result.nfev}\n\n")
        f.write(f"chi-square         = {result.chisqr: .5e}\n")
        f.write(f"reduced chi-square = {result.redchi: .5e}\n")
        f.write(f"chisq_p-value      = {chi2_p_value: .5e} # Chi-squared test\n")
        f.write(
            "ks_p-value         = {: .5e} # Kolmogorov-Smirnov test\n\n".format(
                ks_p_value
            )
        )
        f.write(f"Akaike Information Criterion = {result.aic: .5e}\n")
        f.write(f"Bayesian Information Criterion = {result.bic: .5e}\n")



"""
15N - N-H TROSY CPMG with 0013 phase cycle
==========================================

Analyzes 15N constant-time TROSY CPMG relaxation dispersion experiments for
measurement of ΔD NH in protein systems undergoing millisecond-time-scale
exchange dynamics. Resulting magnetization intensity after the CPMG block is
calculated using the (6n)x(6n), two spin matrix, where n is the number of
states:

[ Nx(a), Ny(a), Nz(a), 2HzNx(a), 2HzNy(a), 2HzNz(a),
  Nx(b), Ny(b), Nz(b), 2HzNx(b), 2HzNy(b), 2HzNz(b),
  ... ]


Note
----

This version is modified such that CPMG pulses are applied with [0013] phase
cycle.  Off resonance effects are taken into account.


Reference
---------

Proc Natl Acad Sci USA (2007) 104, 18473-7


Experimental parameters
-----------------------

  * h_larmor_frq (1H Larmor frequency, in MHz)
  * temperature  (sample temperature, in Celsius)
  * p_total      (optional: protein concentration, in M)
  * l_total      (optional: ligand concentration, in M)
  * time_t2      (CPMG relaxation delay in seconds)
  * carrier      (position of the 15N carrier during the CPMG period, in ppm)
  * time_t2      (CPMG relaxation delay in seconds)
  * pw90         (15N 90 degree pulse width of CPMG pulses, in seconds)
  * time_equil   (equilibration delay at the end of the CPMG period, in seconds [0.0])
  * taub         (p-element delay = 1/4J, in seconds [2.68e-3])
  * ncyc_max     (maximum number of cycles)
  * antitrosy    (perform anti-trosy CPMG RD experiment [False])
  * s3e          (s3e trosy selection [True])


Extra parameters
----------------

  * path         (directory of the profiles)
  * error        (= 'file': uncertainties are taken from the profile files
                  = 'auto': uncertainties are calculated from duplicates)


"""
from functools import reduce

import numpy as np

from chemex.experiments.cpmg import cpmg_profile
from chemex.spindynamics import basis
from chemex.spindynamics import default

EXP_DETAILS = {
    "carrier": {"type": float},
    "time_t2": {"type": float},
    "pw90": {"type": float},
    "taub": {"default": 2.68e-3, "type": float},
    "time_equil": {"default": 0.0, "type": float},
    "ncyc_max": {"type": int},
    "antitrosy": {"default": "False", "type": str},
    "s3e": {"default": "True", "type": str},
}

CP_PHASES = [
    [1, 1, 0, 2, 1, 1, 2, 0, 1, 1, 2, 0, 1, 1, 0, 2],
    [0, 2, 3, 3, 2, 0, 3, 3, 2, 0, 3, 3, 0, 2, 3, 3],
    [0, 0, 1, 3, 0, 0, 3, 1, 0, 0, 3, 1, 0, 0, 1, 3],
    [1, 3, 2, 2, 3, 1, 2, 2, 3, 1, 2, 2, 1, 3, 2, 2],
]


class Profile(cpmg_profile.CPMGProfile):
    """TODO: class docstring."""

    def __init__(self, name, measurements, exp_details, model):

        super().__init__(name, measurements, exp_details, model)

        self.exp_details = self.check_exp_details(exp_details, expected=EXP_DETAILS)

        self.time_t2 = self.exp_details["time_t2"]
        self.time_eq = self.exp_details["time_equil"]
        self.pw90 = self.exp_details["pw90"]
        self.taub = self.exp_details["taub"]
        self.ncyc_max = self.exp_details["ncyc_max"]
        self.antitrosy = self.get_bool(self.exp_details["antitrosy"])
        self.s3e = self.get_bool(self.exp_details["s3e"])

        # Set the liouvillian
        self.liouv = basis.Liouvillian(
            system="ixyzsz",
            state_nb=self.model.state_nb,
            atoms=self.peak.atoms,
            h_larmor_frq=self.conditions["h_larmor_frq"],
            equilibrium=False,
        )

        self.liouv.carrier_i = self.exp_details["carrier"]
        self.liouv.w1_i = 2.0 * np.pi / (4.0 * self.pw90)

        # Set the row vector for detection
        if self.antitrosy:
            self.detect = self.liouv.detect["2izsz_a"] + self.liouv.detect["iz_a"]
        else:
            self.detect = self.liouv.detect["2izsz_a"] - self.liouv.detect["iz_a"]

        # Set the delays in the experiments
        ncycs = self.ncycs[~self.reference]
        self.tau_cps = dict(zip(ncycs, self.time_t2 / (4.0 * ncycs) - 0.75 * self.pw90))
        self.deltas = dict(zip(ncycs, 0.5 * self.pw90 * (self.ncyc_max - ncycs)))
        self.deltas[0] = 0.5 * self.pw90 * (self.ncyc_max - 1)
        self.t_neg = -2.0 * self.pw90 / np.pi
        self.delays = (
            [self.t_neg, self.time_eq, self.taub]
            + list(self.tau_cps.values())
            + list(self.deltas.values())
        )

        # Set the phase cycling of the cpmg pulses
        self.phases = _get_phases(self.ncycs[self.ncycs != 0])

        # Get the parameters this profile depends on
        self.map_names, self.params = default.create_params(
            basis=self.liouv,
            model=self.model,
            nuclei=self.peak.names,
            conditions=self.conditions,
            nh_constraints=True,
        )

        for name, full_name in self.map_names.items():
            if name.startswith(("dw", "r2_i_a")):
                self.params[full_name].set(vary=True)

    def calculate_unscaled_profile(self, **parvals):
        """TODO: Write docstring"""

        self.liouv.update(**parvals)

        # Calculation of the propagators corresponding to all the delays
        delays = dict(zip(self.delays, self.liouv.delays(self.delays)))
        d_taub = delays[self.taub]
        d_neg = delays[self.t_neg]
        d_eq = delays[self.time_eq]
        delta = {ncyc: delays[delay] for ncyc, delay in self.deltas.items()}

        # Calculation of the propagators corresponding to all the pulses
        pulses = self.liouv.pulses_90_180_i()
        p90 = np.array([pulses[name] for name in ["90px", "90py", "90mx", "90my"]])
        p180 = np.array([pulses[name] for name in ["180px", "180py", "180mx", "180my"]])
        p180_sx = self.liouv.perfect180["sx"]

        # Getting the starting magnetization
        mag0 = self._get_mag0(parvals)

        # Calculating the p-element
        if self.antitrosy:
            palmer_ = (
                d_taub @ p90[1] @ p90[0] @ p180_sx @ p90[0] @ p90[1] @ d_taub @ p180_sx
            )
        else:
            palmer_ = (
                d_taub @ p90[0] @ p90[1] @ p180_sx @ p90[1] @ p90[0] @ d_taub @ p180_sx
            )

        palmer = np.mean(p90[[0, 2]] @ palmer_ @ p90[[1, 3]], axis=0)

        # Calculating the cpmg trains
        cp1 = {0: p180[[0, 1]]}
        cp2 = {0: p180[[1, 0]]}

        for ncyc in set(self.ncycs[~self.reference]):

            tau_cp = delays[self.tau_cps[ncyc]]
            phase_cp = self.phases[ncyc]

            echo = tau_cp @ p180 @ tau_cp

            cp_trains = [reduce(np.matmul, echo[phases]) for phases in phase_cp]
            cp_trains = d_neg @ np.array(cp_trains) @ d_neg

            cp1[ncyc] = cp_trains[[0, 1]]
            cp2[ncyc] = cp_trains[[2, 3]]

        # Make profile
        profile = [
            self.liouv.collapse(
                self.detect
                @ d_eq
                @ delta[ncyc]
                @ p90[3]
                @ cp2[ncyc]
                @ palmer
                @ cp1[ncyc]
                @ p90[0]
                @ delta[ncyc]
                @ mag0
            )
            for ncyc in self.ncycs
        ]

        return np.asarray(profile)

    def ncycs_to_nu_cpmgs(self, ncycs=None):
        """Calculate the pulsing frequency, v(CPMG), from ncyc values."""
        if ncycs is None:
            ncycs = self.ncycs

        return ncycs / self.exp_details["time_t2"]

    def _get_mag0(self, parvals):

        mag0 = self.liouv.compute_mag_eq(term="2izsz", **parvals)

        if self.s3e:
            if self.antitrosy:
                mag0 += self.liouv.compute_mag_eq(term="iz", **parvals)
            else:
                mag0 -= self.liouv.compute_mag_eq(term="iz", **parvals)
            mag0 *= 0.5

        return mag0


def _get_phases(ncycs):

    phases = {}

    for ncyc in ncycs:
        phases[ncyc] = np.take(CP_PHASES, np.arange(ncyc), mode="wrap", axis=1)
        phases[ncyc][:2] = np.flip(phases[ncyc][:2], axis=1)

    return phases



"""Unit and regression test for the chemex package."""

import sys

import chemex


def test_chemex_imported():
    """Sample test, will always pass so long as import statement worked"""
    assert "chemex" in sys.modules
